#!/bin/ksh
# File Name:     reorg_online.ksh
#
# Author:        Alex Levy, Sustainable Software Ltd.
#                http://www.sustainablesoftware.net 
#                With grateful thanks to Dan Day for many of the functions!
#
# Purpose:       Reorg database tables and indexes online.
#
# Usage:         reorg.ksh $1 $2 -debug X 
#
# Parameters:    $1 = <instance name> e.g. db2inst1
#                $2 = <database name> e.g. sample
#                -debug X             (where X = Y or N, N is assumed if parameter not supplied)
#
# File dependencies:
#                ${INSTHOME}/sqllib/db2profile
#                reorg.dbadmin.reorgchk_ix.ddl
#                reorg.dbadmin.reorgchk_tb.ddl
#                reorg.pick_index.sql
#                reorg.pick_table.sql
#                reorg.schema.list.${DBNAME}
#
# Notes:         The script REORGs one schema at a time in the instance's schema list. 
#                Within each schema, it REORGs first at table level, then at index level.
#
#                It first drops and recreates 'dbadmin' tables to hold REORGCHK output.
#                For each schema, it then runs the REORGCHK and analyses the output, generates
#                the REORG and RUNSTATS statements using the logic in the 'pick*' scripts
#                and then runs the generated scripts.
#
# Future enhancements:
#                1) Include calls to snapshot_tbreorg - when it's reliable! (PMR 02992,048,866).
#
# Date:          05.08.2005 and enhanced for publication on the web, 20.03.2006
#-------------------------------------------------------------------------------


#-------------------------------------------------------------------------------
# Declare variables.
#
FULL_SCRIPTNAME="$0"                                   # Including path.
SCRIPTNAME="`basename $0`"                             # Excluding path
SHORT_SCRIPTNAME="`basename ${FULL_SCRIPTNAME} .ksh`"  # Strip off path and '.ksh' extension.
PARAMETER_LIST="$*"
DATE=`date +%Y_%m_%d`
REORG_START_TIME=`date`


#-------------------------------------------------------------------------------
# Echo scriptname to standard output.
#
if [[ "$#" -gt 0 ]]; then
   echo "\nScript name:    \"${FULL_SCRIPTNAME}\""
   echo "Parameter list: \"${PARAMETER_LIST}\"\n"
else
   echo "\nScript name:    \"${FULL_SCRIPTNAME}\"\n"
fi


#-------------------------------------------------------------------------------
# Variables to display bold text.
# NB These won't work in cron mail.
#
BOLD=`tput smso`
PLAIN=`tput rmso`


#-------------------------------------------------------------------------------
# Function to print script invocation syntax.
#
SYNTAX()
{
 set "$DEBUG_SWITCH"
 echo "***"
 echo "***  Syntax: I said"
 echo "***          ${SCRIPTNAME} <instance_name> <database_name>"
 echo "***      OR:"
 echo "***          ${SCRIPTNAME} <instance_name> <database_name> -debug X "
 echo "***          (where 'X' is 'Y' or 'N')"
 echo "***"
 echo "***  Option    Purpose"
 echo "***  ------    -------"
 echo "***  -debug    Turns on script debugging."
 echo "***\n"
}


#-------------------------------------------------------------------------------
# Check if input parameters have been supplied.
#
if [[ "$#" -lt 2 ]]; then  # Must supply at least 2 parameters.
   echo "\nERROR - Input parameter(s) not supplied\n"
   SYNTAX  # Call function to print script syntax.
   exit 1  # Terminate script.
else
   typeset -l INSTNAME=${1} # Convert to lower case.
   typeset -l DBNAME=${2}   # Convert to lower case.
fi

DEBUG_SWITCH="+x"  # Default value - disable debugging.

while [[ -n "$3" ]]  # True if parameter has non-zero length.
do
   case "$3" in
   -debug)
      typeset -u DEBUG_OPTION=${4}  # Convert to UPPER case.
      case "$DEBUG_OPTION" in
      Y|YES)
        DEBUG_SWITCH="-x"  # Enable debugging.
        set "$DEBUG_SWITCH"
        ;;
      N|NO)
        DEBUG_SWITCH="+x"  # Disable debugging.
        set "$DEBUG_SWITCH"
        ;;
      *)
        echo "\nERROR - invalid option(s) supplied\n"
        SYNTAX  # Call function to print script syntax.
        exit 2  # Terminate script.
        ;;
      esac
      shift
      ;;
   *)
      echo "\nERROR - invalid parameter(s) supplied\n"
      SYNTAX  # Call function to print script syntax.
      exit 3  # Terminate script.
      ;;
   esac
   shift
done


#-------------------------------------------------------------------------------
# Test that there are no offline backups in progress before attempting a
# connection.
#
CHECK_FOR_BACKUP() 
{
 set "$DEBUG_SWITCH"
 PIDS=`ps -fU "$INSTNAME" | grep -i 'backup' | grep -v 'grep' | awk '{print $2}'`
 if [[ -n "$PIDS" ]]; then
    echo "\n${BOLD}There appears to be a backup in progress:${PLAIN}\n"
    PS_OPTIONS="uw"
    ps "$PS_OPTIONS" | head -1
    for PID in `echo "$PIDS"`
    do
       ps "$PS_OPTIONS" "$PID" | tail -1
    done
    TIDYUP  # Call function to tidy up loose ends.
    exit 4
 else
    # No matches, therefore no backups in progress; take no action.
    : 
 fi
}  


#-------------------------------------------------------------------------------
# Setup DB2 environment variables
#
SETUP_DB2_ENVIRONMENT()
{
 set "$DEBUG_SWITCH"
 echo "\n______________________________________________________________"
 echo "  Setting up DB2 environment"
 echo "____________________________\n"
 PROFILE="${INSTHOME}/sqllib/db2profile"
 FILE_TEST "$PROFILE"
 . "$PROFILE"
 FILE_TEST `which db2`
 echo "DB2 environment setup for:"
 echo "  Instance: "$INSTNAME" "
 echo "  Database: "$DBNAME" "
 echo "______________________________________________________________\n"
}


#-------------------------------------------------------------------------------
# Function to connect to database.
#
CONNECT()
{
 set "$DEBUG_SWITCH"
 echo "\n______________________________________________________________"
 echo "  Connecting to database"
 echo "________________________\n"
 db2 -v "CONNECT TO ${DBNAME}"
 echo "______________________________________________________________\n"
}


#-------------------------------------------------------------------------------
# Function to terminate database connection.
#
TERMINATE()
{
 set "$DEBUG_SWITCH"
 echo "\n______________________________________________________________"
 echo "  Terminating connection to database"
 echo "__________________________________________\n"
 db2 -v "TERMINATE"
 echo "______________________________________________________________\n"
}


#-------------------------------------------------------------------------------
# Function to tidy up loose ends.
#
TIDYUP()
{
 set "$DEBUG_SWITCH"
 if [[ "$1" == "TERMINATE" ]]; then
    TERMINATE  # Call function to terminate database connection.
 fi
 exec 1>&-
 exec 1>&2
 echo "\n--  Reorgs started at: ${REORG_START_TIME}  --" >> "$REORG_LOG"
 echo "--  Reorgs ended   at: `date`  --\n"            >> "$REORG_LOG"
 echo " ----------------------------------------------------------------------------------"
 echo " --"
 echo " -- Contents of reorg log file: ${REORG_LOG}"
 echo " --"
 echo " ----------------------------------------------------------------------------------\n"
 cat "$REORG_LOG"
 echo " ------------------------------------------------------------------------------"
 echo " --"
 echo " --  End of reorg log file: ${REORG_LOG}"
 echo " --"
 echo " ------------------------------------------------------------------------------\n"
}


#-------------------------------------------------------------------------------
# Function to validate directory name based on parameter input and (optionally),
# attempt to create it.
#
DIR_CHK()
{
 set "$DEBUG_SWITCH"
 DIRECTORY="$1"
 if [[ -d "$DIRECTORY" ]]; then
    :                      # Do nothing.
 else                      # If directory doesn't exist ...
    case "$2" in
       create|CREATE)
          mkdir "$DIRECTORY" && chmod 750 "$DIRECTORY"
          RETURN_CODE="$?"
          if [[ "$RETURN_CODE" -ne 0 ]]; then
             echo "ERROR - Unable to create directory: \"${DIRECTORY}\""
          exit 5                  # ... terminate the script.
          fi
          ;;
       *)
          echo "ERROR - Directory doesn't exist: \"${DIRECTORY}\""
          exit 6                  # ... terminate the script.
          ;;
     esac
 fi
}


#-------------------------------------------------------------------------------
# Function to test if script is able to locate a file.
#
FILE_TEST()
{
 set "$DEBUG_SWITCH"
 FILE="$1"
 if [[ -f ${FILE} ]] 
 then :      # file exists and is a regular file, do nothing more.  
 else
    echo "\n********************************************************"
    echo "***"
    echo "***  Unable to locate file: \"${FILE}\""
    echo "***  Exiting script."
    echo "***"
    echo "********************************************************\n"
    if [[ "$2" == "TERMINATE" ]]; then
       TERMINATE  # Call function to terminate database connection.
    fi
    TIDYUP  # Call function to tidy up loose ends.
    exit 7
 fi
}


#-------------------------------------------------------------------------------
# Declare and check directories.
#
INSTHOME="/home/${INSTNAME}"
SCRIPT_DIR="${INSTHOME}/dbascripts"
WORK_DIR="${INSTHOME}/DB2REORG"

DIR_CHK "$INSTHOME"
DIR_CHK "${INSTHOME}/sqllib"
DIR_CHK "$SCRIPT_DIR"
DIR_CHK "$WORK_DIR" "CREATE"


#-------------------------------------------------------------------------------
# Declare and check files.
#
REORG_CHK_TB="${SCRIPT_DIR}/reorg.dbadmin.reorgchk_tb.ddl"
REORG_CHK_IX="${SCRIPT_DIR}/reorg.dbadmin.reorgchk_ix.ddl"
PICK_TABLE="${SCRIPT_DIR}/reorg.pick_table.sql"
PICK_INDEX="${SCRIPT_DIR}/reorg.pick_index.sql"
SCHEMA_LIST="${SCRIPT_DIR}/reorg.schema_list.${DBNAME}"

FILE_TEST "$REORG_CHK_TB"
FILE_TEST "$REORG_CHK_IX"
FILE_TEST "$PICK_TABLE"
FILE_TEST "$PICK_INDEX"
FILE_TEST "$SCHEMA_LIST"


#-------------------------------------------------------------------------------
# Declare output file(s).
# Script prunefiles.ksh takes care of housekeeping the reorg logs.                  
#
REORG_LOG="${WORK_DIR}/${SHORT_SCRIPTNAME}.log.${DBNAME}.${DATE}"
# initialise log file in case of same day re-run.
> ${REORG_LOG}

#-------------------------------------------------------------------------------
# Function to re-direct stdin/stderr to log file.
#
REDIRECT_OUTPUT()
{
 set "$DEBUG_SWITCH"
 echo "\nScript name:    \"${FULL_SCRIPTNAME}\" "        >> "$REORG_LOG"
 echo "Parameter list: \"${PARAMETER_LIST}\" "           >> "$REORG_LOG"
 echo "\n--  Reorgs started at: ${REORG_START_TIME}  --" >> "$REORG_LOG"
 exec 2>&1
 exec 1>>"$REORG_LOG"
}


#-------------------------------------------------------------------------------
# Check the AIX return code for DB2 statements.
# $1 is the return code itself, $2 is the maximum acceptable value.
#
RC_CHECK()
{
 set "$DEBUG_SWITCH"
 RETURN_CODE="$1"
 MAX_RETURN_CODE="$2"
 if [[ "$RETURN_CODE" -gt "$MAX_RETURN_CODE" ]]; then
    echo "Operating system return code from last DB2 statement: " "$RETURN_CODE"
    echo "\n${BOLD}Ever so sorry, please look in: \"${REORG_LOG}\" which is reproduced above:${PLAIN}\n"
    TIDYUP TERMINATE
    exit 1
 fi
}


#-------------------------------------------------------------------------------
# Remove ASAP - set up table first time round without stopping at the first error;
# the script itself drops the control table first.
#
PRIME_PUMP()
{
 set "$DEBUG_SWITCH"
 # Remove ASAP - set up table first time round without stopping at the first error;
 # the script itself drops the control table first.
 db2 -v "CREATE TABLE dbadmin.reorgchk_tb (c1 SMALLINT)" > /dev/null
 db2 -v "CREATE TABLE dbadmin.reorgchk_ix (c1 SMALLINT)" > /dev/null
 db2 -v "COMMIT"                                         > /dev/null
}


#-------------------------------------------------------------------------------
# Function to run REORGs iteratively at schema level.
#
REORG_DATABASE()
{
 set "$DEBUG_SWITCH"
 for SCHEMA in `cat ${SCHEMA_LIST}`
 do
     # Test the schema exists and has tables - sadly,
     # one can't always rely on this in development environments!
     db2 "SELECT 1 FROM sysibm.sysdummy1 \
          WHERE '${SCHEMA}' = ANY(SELECT tabschema FROM syscat.tables)" > /dev/null
     if [[ "$?" -eq 0 ]]
     then
        # Recreate control table to hold REORGCHK stats; note -s switch to stop at first error.
        db2 -t -s -v -f "$REORG_CHK_TB"
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 2
        # Get table REORGCHK stats; run with autocommit off as the procedure creates a DGTT.
        db2 -v +c "CALL sysproc.reorgchk_tb_stats('S','${SCHEMA}')"
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 0
        # Populate control table with said stats
        db2 -v +c "INSERT INTO dbadmin.reorgchk_tb SELECT * FROM session.tb_stats"
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 0
        db2 -v "COMMIT"
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 0
        # Runstats the control table
        db2 -v "RUNSTATS ON TABLE dbadmin.reorgchk_tb ON ALL COLUMNS  \
                WITH DISTRIBUTION ON ALL COLUMNS                      \
                DEFAULT NUM_FREQVALUES 10 NUM_QUANTILES 20            \
                ALLOW READ ACCESS"
        db2 -v "COMMIT"
        # Now select tables in need of a REORG (formulae 1-3), picking the most appropriate index,
        # and generate a reorg script.
        # The -r option does not create empty files, but will append to existing ones,
        # so clear down the script file first.
        # As this SQL is creating a script, there's no need to see the return codes, so I've re-directed
        # command output to /dev/null.

        touch "${WORK_DIR}/${DBNAME}.${SCHEMA}_table_reorg.sql"
        > "${WORK_DIR}/${DBNAME}.${SCHEMA}_table_reorg.sql"
        db2 -t +c -xf "$PICK_TABLE" -r "${WORK_DIR}/${DBNAME}.${SCHEMA}_table_reorg.sql" > /dev/null
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 1
        db2 -v "COMMIT"
        # And run the script if there is one - this script contains RUNSTATS and COMMIT statements
        # as well as the REORGs themselves - hence the db2 -s switch to stop at first error.  
        if [[ -s "${WORK_DIR}/${DBNAME}.${SCHEMA}_table_reorg.sql" ]]
        then 
           db2 -t -s -v -f "${WORK_DIR}/${DBNAME}.${SCHEMA}_table_reorg.sql"
        else
           echo "\n ${BOLD} No table reorgs for schema ${SCHEMA} today ${PLAIN} \n"
        fi
     else
        echo "\n ${BOLD} Table schema ${SCHEMA} has no tables on database ${DBNAME} ${PLAIN} \n"
     fi

     ###
     # Now for the index level REORGs ... 
     # these follow much the same procedural logic as table level REORGs.
     ###
     # Test the schema exists and has indexes - never a certainty in development environments!
     db2 "SELECT 1 FROM sysibm.sysdummy1 \
          WHERE '${SCHEMA}' = ANY(SELECT indschema FROM syscat.indexes)" > /dev/null
     if [[ "$?" -eq 0 ]]
     then
        # Recreate control table to hold REORGCHK stats; note -s switch to stop at first error.
        db2 -t -s -v -f "$REORG_CHK_IX"
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 2
        # Get index REORGCHK stats; run with autocommit off as the procedure creates a DGTT.
        db2 -v +c "CALL sysproc.reorgchk_ix_stats('S','${SCHEMA}')"
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 0
        # Populate control table 
        db2 -v +c "INSERT INTO dbadmin.reorgchk_ix SELECT * FROM session.ix_stats"
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 0
        db2 -v "COMMIT"
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 0
        # Runstats the control table
        db2 -v "RUNSTATS ON TABLE dbadmin.reorgchk_ix ON ALL COLUMNS \
                WITH DISTRIBUTION ON ALL COLUMNS                      \
                DEFAULT NUM_FREQVALUES 10 NUM_QUANTILES 20            \
                ALLOW READ ACCESS"
        db2 -v "COMMIT"
        # Generate a reorg script, based on formulae F4 - F8. 
        # The -r option does not create empty files, but will append to existing ones,
        # so clear down the script file first.
        # As this SQL is creating a script, there's no need to see the return codes, so I've re-directed
        # command output to /dev/null.
        touch "${WORK_DIR}/${DBNAME}.${SCHEMA}_index_reorg.sql"
        > "${WORK_DIR}/${DBNAME}.${SCHEMA}_index_reorg.sql"
        db2 -t +c -x -v -f "$PICK_INDEX" -r "${WORK_DIR}/${DBNAME}.${SCHEMA}_index_reorg.sql" > /dev/null
        RETURN_CODE="$?" && RC_CHECK "$RETURN_CODE" 1
        db2 -v "COMMIT"
        # Like the table REORG script, this one contains interleaved REORG, RUNSTATS and COMMIT 
        # statements.
        if [[ -s "${WORK_DIR}/${DBNAME}.${SCHEMA}_index_reorg.sql" ]]
        then 
           db2 -t -s -v -f "${WORK_DIR}/${DBNAME}.${SCHEMA}_index_reorg.sql"
        else
           echo "\n ${BOLD} No index reorgs for schema ${SCHEMA} today ${PLAIN} \n"
        fi
     else
        echo "\n ${BOLD} Schema ${SCHEMA} has no indexes on database ${DBNAME} ${PLAIN} \n"
     fi
 done
}


#-------------------------------------------------------------------------------
# Main logic to call functions.
#
REDIRECT_OUTPUT        # Call function to re-direct output to log file.
CHECK_FOR_BACKUP       # Exit gracefully if a backup is in progress, else ...
SETUP_DB2_ENVIRONMENT  # Call function to setup DB2 environment.
CONNECT                # Call function to connect to database.
PRIME_PUMP             # Temporary code for first run only.
REORG_DATABASE         # Call function to REORG database, one schema at a time.
TIDYUP                 # Call function to tidy up loose ends.
